﻿////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// XPS Driver SDK magstripe sample.
////////////////////////////////////////////////////////////////////////////////
using System;
using System.IO;
using dxp01sdk;

public class CommandLineOptions {
    public string printerName;
    public bool readMagstripe;
    public bool print;
    public bool jobCompletion;
    public bool jisRequest;
    public bool checkSupplies;
};

class magstripe {
    private static void usage() {
        string thisExeName = Path.GetFileName(Environment.GetCommandLineArgs()[0]);
        Console.WriteLine(thisExeName + " demonstrates interactive mode magnetic stripe encoding with");
        Console.WriteLine("  options to read the magnetic stripe data, print, and poll for job");
        Console.WriteLine("  completion status.");
        Console.WriteLine();
        Console.WriteLine("Uses hardcoded data for magnetic stripe and printing.");
        Console.WriteLine();
        Console.WriteLine(thisExeName + " -n <printername> [-r] [-p] [-c] [-j] [-i] [-c]");
        Console.WriteLine();
        Console.WriteLine("options:");
        Console.WriteLine("  -n <printername>. Required. Try -n \"XPS Card Printer\".");
        Console.WriteLine("  -r Reads back the encoded magnetic stripe data.");
        Console.WriteLine("  -p Print simple black text on front side of card. ");
        Console.WriteLine("  -j JIS magnetic. ");
        Console.WriteLine("  -i Check supplies before printing. Default is to not check.");
        Console.WriteLine("  -c Poll for job completion.");
        Console.WriteLine();
        Console.WriteLine(thisExeName + " -n \"XPS Card Printer\"");
        Console.WriteLine("  Encodes data on all three tracks of an ISO 3-track magnetic stripe on the");
        Console.WriteLine("  backside of a card.");
        Console.WriteLine();
        Console.WriteLine(thisExeName + " -n \"XPS Card Printer\" -r -p -c");
        Console.WriteLine("  Encodes data on all three tracks of an ISO 3-track magnetic stripe on the");
        Console.WriteLine("  backside of a card, reads the data from the magnetic stripe and displays");
        Console.WriteLine("  it, prints black text on the front side of the card, and polls and");
        Console.WriteLine("  displays job status.");
        Environment.Exit(-1);
    }

    static CommandLineOptions GetCommandlineOptions(string[] args) {
        if (args.Length == 0) usage();

        CommandLineOptions commandLineOptions = new CommandLineOptions();
        CommandLine.Utility.Arguments arguments = new CommandLine.Utility.Arguments(args);

        if (!string.IsNullOrEmpty(arguments["h"])) usage();

        if (string.IsNullOrEmpty(arguments["n"])) {
            Console.WriteLine("printer name is required");
            Environment.Exit(-1);
        }
        else {
            // we might have a -n with no printer name:
            bool boolVal = false;
            if (Boolean.TryParse(arguments["n"], out boolVal)) {
                Console.WriteLine("printer name is required");
                Environment.Exit(-1);
            }
            commandLineOptions.printerName = arguments["n"];
        }

        commandLineOptions.readMagstripe = !string.IsNullOrEmpty(arguments["r"]);
        commandLineOptions.print = !string.IsNullOrEmpty(arguments["p"]);
        commandLineOptions.jobCompletion = !string.IsNullOrEmpty(arguments["c"]);
        commandLineOptions.checkSupplies = !string.IsNullOrEmpty(arguments["i"]);
        commandLineOptions.jisRequest = !string.IsNullOrEmpty(arguments["j"]);

        return commandLineOptions;
    }

    static void ResumeJob(BidiSplWrap bidiSpl, int printerJobID, int errorCode) {
        string xmlFormat = strings.PRINTER_ACTION_XML;
        string input = string.Format(xmlFormat, (int) Actions.Resume, printerJobID, errorCode);
        bidiSpl.SetPrinterData(strings.PRINTER_ACTION, input);
    }

    public static void ReadMagstripe(BidiSplWrap bidiSpl, bool jisRequest) {

        // replace schema string MAGSTRIPE_READ to MAGSTRIPE_READ_FRONT for front side read
        string printerStatusXML = bidiSpl.GetPrinterData(strings.MAGSTRIPE_READ);

        PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
        if (0 == printerStatusValues._errorCode) {
            Console.WriteLine("Magstripe data read successfully; printer job id: " + printerStatusValues._printerJobID);

            string track1 = "";
            string track2 = "";
            string track3 = "";

            Util.ParseMagstripeStrings(printerStatusXML, ref track1, ref track2, ref track3, jisRequest);

            if (track1.Length != 0) {
                // Convert the Base64 UUEncoded output.
                byte[] binaryData = System.Convert.FromBase64String(track1);
                string str = System.Text.Encoding.UTF8.GetString(binaryData);
                Console.WriteLine(" track1 Base64 decoded: " + str);
            }

            if (track2.Length != 0) {
                // Convert the Base64 UUEncoded output.
                byte[] binaryData = System.Convert.FromBase64String(track2);
                string str = System.Text.Encoding.UTF8.GetString(binaryData);
                Console.WriteLine(" track2 Base64 decoded: " + str);
            }

            if (track3.Length != 0) {
                // Convert the Base64 UUEncoded output.
                byte[] binaryData = System.Convert.FromBase64String(track3);
                string str = System.Text.Encoding.UTF8.GetString(binaryData);
                Console.WriteLine(" track3 Base64 decoded: " + str);
            }
        }
        else {
            throw new BidiException(printerStatusValues._errorString, printerStatusValues._printerJobID, printerStatusValues._errorCode);
        }
    }

    public static void EncodeMagstripe(BidiSplWrap bidiSpl, bool jisRequest) {

        // Hardcoded XML to encode all 3 tracks in IAT mode.
        // track 1 = "TRACK1", track 2 = "1122", track 3 = "321"
        string trackDataXML = "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
            + "<magstripe>"
            + "<track number=\"1\"><base64Data>VFJBQ0sx</base64Data></track>"
            + "<track number=\"2\"><base64Data>MTEyMg==</base64Data></track>"
            + "<track number=\"3\"><base64Data>MzIx</base64Data></track>"
            + "</magstripe>";

        if (jisRequest) {
            // JIS only allows track 3 = "321"
            trackDataXML = "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
            + "<magstripe>"
            + "<track number=\"1\"><base64Data></base64Data></track>"
            + "<track number=\"2\"><base64Data></base64Data></track>"
            + "<track number=\"3\"><base64Data>MzIx</base64Data></track>"
            + "</magstripe>";
        }

        // replace schema string MAGSTRIPE_ENCODE to MAGSTRIPE_ENCODE_FRONT for front side encode
        string printerStatusXML = bidiSpl.SetPrinterData(strings.MAGSTRIPE_ENCODE, trackDataXML);
        PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);

        if (0 != printerStatusValues._errorCode) {
            throw new BidiException(printerStatusValues._errorString, printerStatusValues._printerJobID, printerStatusValues._errorCode);
        }
    }

    ////////////////////////////////////////////////////////////////////////////////
    // Main()
    ////////////////////////////////////////////////////////////////////////////////
    static void Main(string[] args) {

        CommandLineOptions commandLineOptions = GetCommandlineOptions(args);

        BidiSplWrap bidiSpl = null;
        int printerJobID = 0;

        try {

            bidiSpl = new BidiSplWrap();
            bidiSpl.BindDevice(commandLineOptions.printerName);

            string driverVersionXml = bidiSpl.GetPrinterData(strings.SDK_VERSION);
            Console.WriteLine(Environment.NewLine + "driver version: " + Util.ParseDriverVersionXML(driverVersionXml) + Environment.NewLine);

            string printerOptionsXML = bidiSpl.GetPrinterData(strings.PRINTER_OPTIONS2);
            PrinterOptionsValues printerOptionsValues = Util.ParsePrinterOptionsXML(printerOptionsXML);

            if ("Ready" != printerOptionsValues._printerStatus && "Busy" != printerOptionsValues._printerStatus) {
                throw new Exception(commandLineOptions.printerName + " is not ready. status: " + printerOptionsValues._printerStatus);
            }

            string hopperID = string.Empty;
            printerJobID = Util.StartJob(
                bidiSpl,
                commandLineOptions.checkSupplies,
                hopperID);

            EncodeMagstripe(bidiSpl, commandLineOptions.jisRequest);

            if (commandLineOptions.readMagstripe) {
                ReadMagstripe(bidiSpl, commandLineOptions.jisRequest);
            }

            // Check if user wants print
            if (commandLineOptions.print) {
                // this also waits for the print spooling to finish:
                Util.PrintTextAndGraphics(bidiSpl, commandLineOptions.printerName);
            }

            bidiSpl.SetPrinterData(strings.ENDJOB);

            if (commandLineOptions.jobCompletion) {
                Util.PollForJobCompletion(bidiSpl, printerJobID);
            }
        }
        catch (BidiException e) {
            Console.WriteLine(e.Message);
            Util.CancelJob(bidiSpl, e.PrinterJobID, e.ErrorCode);
        }
        catch (Exception e) {
            Console.WriteLine(e.Message);
            if (0 != printerJobID) {
                Util.CancelJob(bidiSpl, printerJobID, 0);
            }
        }
        finally {
            bidiSpl.UnbindDevice();
        }
    }
}