////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
////////////////////////////////////////////////////////////////////////////////
package com.datacard.xpsprinter.sdk.examples;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.Line2D;
import java.awt.image.BufferedImage;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterJob;
import java.net.URL;
import javax.imageio.ImageIO;
import javax.print.PrintService;
import javax.print.PrintServiceLookup;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.Copies;
import javax.print.attribute.standard.MediaPrintableArea;
import javax.print.attribute.standard.MediaSize;
import javax.print.attribute.standard.OrientationRequested;
import javax.print.attribute.standard.Sides;

public class JavaPrint implements Printable {

    private static final long serialVersionUID = 1L;
    private final double POINTS_PER_INCH = 72;
    private String mPrinterName = null;
    private boolean mDuplex;
    private boolean mLandscape;
    private boolean mDoBacksideEncode;
    private int mCopyCount;
    private boolean mBlockPrintMagstripe;
    private boolean mBlockPrintSmartcard;
    private boolean mJISMagstripe;

    public JavaPrint(String aPrinterName) {
        mPrinterName = aPrinterName;
        mDuplex = false;
        mDoBacksideEncode = false;
        mBlockPrintSmartcard = false;
        mBlockPrintMagstripe = false;
        mJISMagstripe = false;
        mCopyCount = 1;
    }

    public JavaPrint(String aPrinterName, boolean isDuplex, boolean isLandscape,
            boolean doBacksideEscapeEncode, boolean isBlockPrintMagstripe,
            boolean isBlockPrintSmartcard, int copyCount) {
        mPrinterName = aPrinterName;
        mDuplex = isDuplex;
        mLandscape = isLandscape;
        mDoBacksideEncode = doBacksideEscapeEncode;
        mBlockPrintSmartcard = isBlockPrintSmartcard;
        mBlockPrintMagstripe = isBlockPrintMagstripe;
        mJISMagstripe = false;
        mCopyCount = copyCount;
    }

    public JavaPrint(String aPrinterName, boolean isDuplex, boolean isLandscape,
            boolean doBacksideEscapeEncode, boolean isBlockPrintMagstripe,
            boolean isBlockPrintSmartcard, boolean isJISMagstripe, int copyCount) {
        mPrinterName = aPrinterName;
        mDuplex = isDuplex;
        mLandscape = isLandscape;
        mDoBacksideEncode = doBacksideEscapeEncode;
        mBlockPrintSmartcard = isBlockPrintSmartcard;
        mBlockPrintMagstripe = isBlockPrintMagstripe;
        mJISMagstripe = isJISMagstripe;
        mCopyCount = copyCount;
    }

    private PrintService GetPrinterService() {
        PrintService[] printServices = PrintServiceLookup.lookupPrintServices(
                null, null);

        for (PrintService printer : printServices) {
            if (printer.getName().compareToIgnoreCase(mPrinterName) == 0) {
                System.out.println("Found Printer: " + printer.getName());
                return printer;
            }
        }
        System.out.println("Did not match any printer name as : " + mPrinterName);
        return null;
    }

    public void PrintDemo() {
        PrintService printService = GetPrinterService();
        if (printService == null) {
            return;
        }

        try {
            PrinterJob printJob = PrinterJob.getPrinterJob();
            printJob.setPrintService(printService);

            // Set the printable class to this one since we are implementing the
            // Printable interface:
            PrintRequestAttributeSet aset = new HashPrintRequestAttributeSet();

            if (mLandscape) {
                aset.add(OrientationRequested.LANDSCAPE);
            } else {
                aset.add(OrientationRequested.PORTRAIT);
            }
            aset.add(new MediaPrintableArea(0.0f, 0.0f, 53.98f, 85.6f,
                    MediaSize.MM));
            aset.add(new Copies(mCopyCount));

            if (mDuplex) {
                aset.add(Sides.DUPLEX);
                mDuplex = true;
            } else {
                aset.add(Sides.ONE_SIDED);
            }

            printJob.setPrintable(this);
            printJob.print(aset);

        } catch (Exception PrintException) {
            PrintException.printStackTrace();
        }
    }

    /**
     * Method: print
     * <p>
     *
     * This class is responsible for rendering a page using the provided
     * parameters. The result will be a grid where each cell is one half inch
     * square.
     *
     * @param g a value of type Graphics
     * @param pageFormat a value of type PageFormat
     * @param pageNumber a value of type int
     * @return an int error code.
     */
    @Override
    public int print(Graphics g, PageFormat pageFormat, int pageNumber) {
        int imgStartX = 10;
        int imgStartY = 10;

        // --- Validate the page number, we only print max two pages
        if (pageNumber > 1) {
            System.out.println("Print job has been sent to spooler.");
            return Printable.NO_SUCH_PAGE;
        }

        // --- Translate the origin to be (0,0)
        Graphics2D graphics2D = (Graphics2D) g;
        graphics2D.setFont(new Font(null, Font.PLAIN, 8));

        graphics2D.translate(pageFormat.getImageableX(),
                pageFormat.getImageableY());

        final int printerDPI = 300;

        double pageWidth = pageFormat.getWidth() * printerDPI / POINTS_PER_INCH;
        double pageHeight = pageFormat.getHeight() * printerDPI / POINTS_PER_INCH;

        if (pageNumber == 0) {
            graphics2D.setColor(Color.black);
            Line2D.Double line = new Line2D.Double();

            // --- Print the vertical lines
            for (int lineIndex = 0; lineIndex < pageWidth; lineIndex += POINTS_PER_INCH / 2) {
                line.setLine(lineIndex, 0, lineIndex, pageHeight);
                graphics2D.draw(line);
                graphics2D.drawString("V" + Integer.toString(lineIndex), lineIndex,
                        20);
            }

            // --- Print the horizontal lines
            for (int lineIndex = 0; lineIndex < pageHeight; lineIndex += POINTS_PER_INCH / 2) {
                line.setLine(0, lineIndex, pageWidth, lineIndex);
                graphics2D.draw(line);
                graphics2D.drawString("H" + Integer.toString(lineIndex), 0,
                        lineIndex);
            }

            // draw a color bitmap
            drawImage(graphics2D, pageFormat, "images/color.jpg", imgStartX,
                    imgStartY, (int) pageWidth, (int) pageHeight);

            graphics2D.drawString("Front side", (int) 60, (int) 100);

            if (mBlockPrintSmartcard) {
                blockPrintSmartcard(graphics2D);
            }

            return (Printable.PAGE_EXISTS);
        } else if (pageNumber == 1 && mDuplex) {
            imgStartX = 72;
            imgStartY = 72;

            graphics2D = (Graphics2D) g;
            graphics2D.setColor(Color.black);

            // back side escape encode
            if (mDoBacksideEncode) {
                WriteMagstripeEscapes(graphics2D);
            }

            // draw a 1 bit-per-pixel bitmap to the KPanel:
            drawImage(graphics2D, pageFormat, "images/House_1bpp.png", imgStartX,
                    imgStartY, (int) pageWidth, (int) pageHeight);

            graphics2D.drawString("back side ", (int) 72, (int) 44);

            if (mBlockPrintMagstripe) {
                blockPrintMagstripe(graphics2D);
            }

            return (Printable.PAGE_EXISTS);
        } else {
            return (NO_SUCH_PAGE);
        }
    }

    private void WriteJISMagstripeEscapes(Graphics2D g2d) {
        String track3Escape = "1234567890ｱｴｲｵｳｱｴｲｵｳｱｴｲｵｳ1234567890";

        Font oldFont = g2d.getFont();

        // set JIS font
        g2d.setFont(new Font("Mag Track JIS XPS Card Printer", Font.PLAIN, 4));
        g2d.drawString(track3Escape, 20, 40);

        // restore the old font
        g2d.setFont(oldFont);
    }

    private void WriteMagstripeEscapes(Graphics2D g2d) {

        if (mJISMagstripe) {
            WriteJISMagstripeEscapes(g2d);
        } else {
            final String track1Escape = "~1ABC 123";
            final String track2Escape = "~2456";
            final String track3Escape = "~3789";

            g2d.drawString(track1Escape, 0, 0);
            g2d.drawString(track2Escape, 0, 0);
            g2d.drawString(track3Escape, 0, 0);
        }
    }

    private void blockPrintSmartcard(Graphics2D g2d) {
        System.out.println("BlockPrintSmartcard");
        g2d.drawString("~TR% 7 21 16 16?", 0, 30);
        g2d.drawString("~PB% 7 21 16 16?", 0, 30);
    }

    private void blockPrintMagstripe(Graphics g2d) {
        if (!mJISMagstripe) {
            g2d.drawString("~TR% 0 35 10 120?", 0, 0);
            g2d.drawString("~PB% 0 35 10 120?", 0, 0);
        } else {
            g2d.drawString("~TR% 0 35 3.33 120?", 0, 0);
            g2d.drawString("~PB% 0 35 3.33 120?", 0, 0);
        }
    }

    private void drawImage(Graphics2D graphics2D, PageFormat pageFormat,
            String aFileName, int x, int y, int pageWidth, int pageHeight) {

        BufferedImage img = getSystemImage(aFileName);

        if (img == null) {
            return;
        }

        int w = img.getWidth();
        int h = img.getHeight();
        int destW = (int) (w * 0.3) + x;
        int destH = (int) (h * 0.3) + y;

        if (w > pageWidth) {
            destW = pageWidth;
        }

        if (h > pageHeight) {
            destH = pageHeight;
        }

        graphics2D.drawImage(img, x, y, destW, destH, 0, 0, w, h, null);
    }

    private BufferedImage getSystemImage(String filename) {
        if ((filename == null) || (filename.length() == 0)) {
            return null;
        }

        URL url = ClassLoader.getSystemResource(filename);
        if (url == null) {
            System.out.println("Unable to locate file: " + filename);
            return null;
        }

        try {
            return ImageIO.read(ClassLoader.getSystemResourceAsStream(filename));
        } catch (Exception e) {
            System.out.println(e.getMessage());
            return null;
        }
    }
}
